@@ -22,7 +22,6 @@ import atexit
 import ConfigParser
 import errno
 import grp
-import hashlib
 import json
 import logging
 import os
@@ -292,7 +291,21 @@ class RpmHelper(object):
         return command.status == 0
 
     @classmethod
-    def install(cls, packages, rpms=True):
+    def zypper_package_available(cls, pkg):
+        """Indicates whether pkg is available via zypper.
+
+        Arguments:
+            pkg -- A package name (with optional version and release spec).
+                   e.g., httpd
+                   e.g., httpd-2.2.22
+                   e.g., httpd-2.2.22-1.fc16
+        """
+        cmd_str = "zypper -n --no-refresh search %s" % pkg
+        command = CommandRunner(cmd_str).run()
+        return command.status == 0
+
+    @classmethod
+    def install(cls, packages, rpms=True, zypper=False):
         """Installs (or upgrades) a set of packages via RPM or via Yum.
 
         Arguments:
@@ -312,6 +325,10 @@ class RpmHelper(object):
             cmd = "rpm -U --force --nosignature "
             cmd += " ".join(packages)
             LOG.info("Installing packages: %s" % cmd)
+        elif zypper:
+            cmd = "zypper -n install "
+            cmd += " ".join(packages)
+            LOG.info("Installing packages: %s" % cmd)
         else:
             cmd = "yum -y install "
             cmd += " ".join(packages)
@@ -321,7 +338,7 @@ class RpmHelper(object):
             LOG.warn("Failed to install packages: %s" % cmd)
 
     @classmethod
-    def downgrade(cls, packages, rpms=True):
+    def downgrade(cls, packages, rpms=True, zypper=False):
         """Downgrades a set of packages via RPM or via Yum.
 
         Arguments:
@@ -338,6 +355,13 @@ class RpmHelper(object):
         """
         if rpms:
             cls.install(packages)
+        elif zypper:
+            cmd = "zypper -n install --oldpackage "
+            cmd += " ".join(packages)
+            LOG.info("Downgrading packages: %s", cmd)
+            command = CommandRunner(cmd).run()
+            if command.status:
+                LOG.warn("Failed to downgrade packages: %s" % cmd)
         else:
             cmd = "yum -y downgrade "
             cmd += " ".join(packages)
@@ -391,6 +415,51 @@ class PackagesHandler(object):
             cmd_str = 'easy_install %s' % (pkg_name)
             CommandRunner(cmd_str).run()
 
+    def _handle_zypper_packages(self, packages):
+        """Handle installation, upgrade, or downgrade of packages via yum.
+
+        Arguments:
+        packages -- a package entries map of the form:
+                      "pkg_name" : "version",
+                      "pkg_name" : ["v1", "v2"],
+                      "pkg_name" : []
+
+        For each package entry:
+          * if no version is supplied and the package is already installed, do
+            nothing
+          * if no version is supplied and the package is _not_ already
+            installed, install it
+          * if a version string is supplied, and the package is already
+            installed, determine whether to downgrade or upgrade (or do nothing
+            if version matches installed package)
+          * if a version array is supplied, choose the highest version from the
+            array and follow same logic for version string above
+        """
+        # collect pkgs for batch processing at end
+        installs = []
+        downgrades = []
+        for pkg_name, versions in packages.iteritems():
+            ver = RpmHelper.newest_rpm_version(versions)
+            pkg = "%s-%s" % (pkg_name, ver) if ver else pkg_name
+            if RpmHelper.rpm_package_installed(pkg):
+                # FIXME:print non-error, but skipping pkg
+                pass
+            elif not RpmHelper.zypper_package_available(pkg):
+                LOG.warn("Skipping package '%s' - unavailable via zypper", pkg)
+            elif not ver:
+                installs.append(pkg)
+            else:
+                current_ver = RpmHelper.rpm_package_version(pkg)
+                rc = RpmHelper.compare_rpm_versions(current_ver, ver)
+                if rc < 0:
+                    installs.append(pkg)
+                elif rc > 0:
+                    downgrades.append(pkg)
+        if installs:
+            RpmHelper.install(installs, rpms=False, zypper=True)
+        if downgrades:
+            RpmHelper.downgrade(downgrades, zypper=True)
+
     def _handle_yum_packages(self, packages):
         """Handle installation, upgrade, or downgrade of packages via yum.
 
@@ -462,6 +531,7 @@ class PackagesHandler(object):
 
     # map of function pointers to handle different package managers
     _package_handlers = {"yum": _handle_yum_packages,
+                         "zypper": _handle_zypper_packages,
                          "rpm": _handle_rpm_packages,
                          "apt": _handle_apt_packages,
                          "rubygems": _handle_gem_packages,
@@ -522,7 +592,7 @@ class FilesHandler(object):
                     f.write(json.dumps(meta['content'], indent=4))
                     f.close()
             elif 'source' in meta:
-                CommandRunner('wget -O %s %s' % (dest, meta['source'])).run()
+                CommandRunner('curl -o %s %s' % (dest, meta['source'])).run()
             else:
                 LOG.error('%s %s' % (dest, str(meta)))
                 continue
@@ -601,20 +671,20 @@ class SourcesHandler(object):
         basename = os.path.basename(url)
         stype = self._source_type(url)
         if stype == '.tgz':
-            cmd = "wget -q -O - '%s' | gunzip | tar -xvf -" % url
+            cmd = "curl -s '%s' | gunzip | tar -xvf -" % url
         elif stype == '.tbz2':
-            cmd = "wget -q -O - '%s' | bunzip2 | tar -xvf -" % url
+            cmd = "curl -s '%s' | bunzip2 | tar -xvf -" % url
         elif stype == '.zip':
             tmp = self._url_to_tmp_filename(url)
-            cmd = "wget -q -O '%s' '%s' && unzip -o '%s'" % (tmp, url, tmp)
+            cmd = "curl -s -o '%s' '%s' && unzip -o '%s'" % (tmp, url, tmp)
         elif stype == '.tar':
-            cmd = "wget -q -O - '%s' | tar -xvf -" % url
+            cmd = "curl -s '%s' | tar -xvf -" % url
         elif stype == '.gz':
             (r, ext) = self._splitext(basename)
-            cmd = "wget -q -O - '%s' | gunzip > '%s'" % (url, r)
+            cmd = "curl -s '%s' | gunzip > '%s'" % (url, r)
         elif stype == '.bz2':
             (r, ext) = self._splitext(basename)
-            cmd = "wget -q -O - '%s' | bunzip2 > '%s'" % (url, r)
+            cmd = "curl -s '%s' | bunzip2 > '%s'" % (url, r)
 
         if cmd != '':
             cmd = "mkdir -p '%s'; cd '%s'; %s" % (dest, dest, cmd)
@@ -643,37 +713,47 @@ class ServicesHandler(object):
         self.hooks = hooks
 
     def _handle_sysv_command(self, service, command):
-        service_exe = "/sbin/service"
-        enable_exe = "/sbin/chkconfig"
-        cmd = ""
-        if "enable" == command:
-            cmd = "%s %s on" % (enable_exe, service)
-        elif "disable" == command:
-            cmd = "%s %s off" % (enable_exe, service)
-        elif "start" == command:
-            cmd = "%s %s start" % (service_exe, service)
-        elif "stop" == command:
-            cmd = "%s %s stop" % (service_exe, service)
-        elif "status" == command:
-            cmd = "%s %s status" % (service_exe, service)
-        command = CommandRunner(cmd)
-        command.run()
-        return command
+        if os.path.exists("/bin/systemctl"):
+            service_exe = "/bin/systemctl"
+            service = '%s.service' % service
+            service_start = '%s start %s'
+            service_status = '%s status %s'
+            service_stop = '%s stop %s'
+        elif os.path.exists("/sbin/service"):
+            service_exe = "/sbin/service"
+            service_start = '%s %s start'
+            service_status = '%s %s status'
+            service_stop = '%s %s stop'
+        else:
+            service_exe = "/usr/sbin/service"
+            service_start = '%s %s start'
+            service_status = '%s %s status'
+            service_stop = '%s %s stop'
+
+        if os.path.exists("/bin/systemctl"):
+            enable_exe = "/bin/systemctl"
+            enable_on = '%s enable %s'
+            enable_off = '%s disable %s'
+        elif os.path.exists("/sbin/chkconfig"):
+            enable_exe = "/sbin/chkconfig"
+            enable_on = '%s %s on'
+            enable_off = '%s %s off'
+        else:
+            enable_exe = "/usr/sbin/update-rc.d"
+            enable_on = '%s %s enable'
+            enable_off = '%s %s disable'
 
-    def _handle_systemd_command(self, service, command):
-        exe = "/bin/systemctl"
         cmd = ""
-        service = '%s.service' % service
         if "enable" == command:
-            cmd = "%s enable %s" % (exe, service)
+            cmd = enable_on % (enable_exe, service)
         elif "disable" == command:
-            cmd = "%s disable %s" % (exe, service)
+            cmd = enable_off % (enable_exe, service)
         elif "start" == command:
-            cmd = "%s start %s" % (exe, service)
+            cmd = service_start % (service_exe, service)
         elif "stop" == command:
-            cmd = "%s stop %s" % (exe, service)
+            cmd = service_stop % (service_exe, service)
         elif "status" == command:
-            cmd = "%s status %s" % (exe, service)
+            cmd = service_status % (service_exe, service)
         command = CommandRunner(cmd)
         command.run()
         return command
@@ -725,7 +805,7 @@ class ServicesHandler(object):
     # map of function pointers to various service handlers
     _service_handlers = {
         "sysvinit": _handle_sysv_command,
-        "systemd": _handle_systemd_command
+        "systemd": _handle_sysv_command
     }
 
     def _service_handler(self, manager_name):
@@ -1069,7 +1149,7 @@ class Metadata(object):
 
         url = 'http://169.254.169.254/openstack/2012-08-10/meta_data.json'
         if not os.path.exists(cache_path):
-            CommandRunner('wget -O %s %s' % (cache_path, url)).run()
+            CommandRunner('curl -o %s %s' % (cache_path, url)).run()
         try:
             with open(cache_path) as fd:
                 try:
@@ -1108,6 +1188,11 @@ class Metadata(object):
                True -- success
               False -- error
         """
+        if self.resource is not None:
+            res_last_path = last_path + '_' + self.resource
+        else:
+            res_last_path = last_path
+
         if meta_str:
             self._data = meta_str
         else:
@@ -1127,7 +1212,7 @@ class Metadata(object):
                 # cached metadata or the logic below could re-run a stale
                 # cfn-init-data
                 fd = None
-                for filepath in [last_path, default_path]:
+                for filepath in [res_last_path, last_path, default_path]:
                     try:
                         fd = open(filepath)
                     except IOError:
@@ -1150,18 +1235,21 @@ class Metadata(object):
         else:
             self._metadata = self._data
 
-        cm = hashlib.md5(json.dumps(self._metadata))
-        current_md5 = cm.hexdigest()
-        old_md5 = None
+        last_data = ""
+        for metadata_file in [res_last_path, last_path]:
+            try:
+                with open(metadata_file) as lm:
+                    try:
+                        last_data = json.load(lm)
+                    except ValueError:
+                        pass
+                    lm.close()
+            except IOError:
+                LOG.warn("Unable to open local metadata : %s" %
+                         metadata_file)
+                continue
 
-        try:
-            with open(last_path) as lm:
-                om = hashlib.md5()
-                om.update(lm.read())
-                old_md5 = om.hexdigest()
-        except Exception:
-            pass
-        if old_md5 != current_md5:
+        if self._metadata != last_data:
             self._has_changed = True
 
         # if cache dir does not exist try to create it
@@ -1181,6 +1269,9 @@ class Metadata(object):
             os.chmod(cf.name, 0o600)
             cf.write(json.dumps(self._metadata))
             os.rename(cf.name, last_path)
+            cf.close()
+            if res_last_path != last_path:
+                shutil.copy(last_path, res_last_path)
 
         return True
 
